// dict.c
// vi:set ts=3

#include <PalmOS.h>
#include <zlib.h>

#include "dictrc.h"
#include "dict.h"

/* Global variables */
char DictDBName[dmDBNameLength];
UInt16 ViewPage;
char ViewTitle[64];
char buffer[DE_BUFFER_SIZE];
GlobalData gd;
DmOpenRef CurrentDB, StatusDB;
FormPtr mainform;
char quit = 0;

UInt32  PilotMain (UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
	UInt16 error = 0;

	switch (cmd)
	{

	case sysAppLaunchCmdNormalLaunch:

		if (( error = StartApplication() ) != 0 ) { // Application start code
			return ( error );
		}

		EventLoop();	// Event loop

		StopApplication();	// Application stop code
		break;

	}
	return ( error );
}

Err StartApplication(void)
{
	Err err;
	DmSearchStateType dbState;	
	UInt16 cardNo;
	LocalID dbID;
	UInt16 count = 0;

	MemHandle rechandle;
	MemPtr recptr;

	LocalID StatusID;
	UInt16 StatusCardNo;
	Boolean createStatus = true;

	/* Count how many databases there are */
	if( DmGetNextDatabaseByTypeCreator( true, &dbState, DictType, DictCreator, false, &cardNo, &dbID ) == 0) {
		do { 
			count++;
		} while( DmGetNextDatabaseByTypeCreator( false, &dbState, DictType, DictCreator, false, &cardNo, &dbID ) == 0);
	}

	if( count == 0 ) {
		/* There are no databases */
		FrmCustomAlert(AlertForm,"Cannot find dictionary's database.",NULL,NULL);
		quit = 1;
		return 1;
	}

	MemSet(&gd,sizeof(GlobalData),0);

	CurrentDB = DmOpenDatabaseByTypeCreator( DictType, DictCreator, dmModeReadOnly );

	if ((StatusDB = DmOpenDatabaseByTypeCreator(DictType,DictStatCreator,dmModeReadWrite))) {
		if (DmNumRecords(StatusDB)<2) {
			// delete db
			DmOpenDatabaseInfo(StatusDB, &StatusID, NULL, NULL, &StatusCardNo, NULL);
			DmCloseDatabase(StatusDB);
			DmDeleteDatabase(StatusCardNo, StatusID);
		}
		else {
			// status exist, load state
			rechandle = DmGetRecord(StatusDB, 0);
			recptr = MemHandleLock(rechandle);
			MemMove(&gd, recptr, sizeof(GlobalData));
			MemHandleUnlock(rechandle);
			DmReleaseRecord(StatusDB, 0, false);
			createStatus = false;
		}
	}
	if (createStatus) {
		FormPtr frm = FrmInitForm(AboutForm);
		char oldtitle[80], title[80];
		UInt16 butClose = FrmGetObjectIndex(frm, AboutCloseButton);

		FrmHideObject(frm, butClose);
		StrCopy(oldtitle,FrmGetTitle(frm));
		StrCopy(title,"Loading...");
		FrmSetTitle(frm, title);
		FrmDrawForm(frm);

		gd.de_word_record = gd.de_mean_record = 255;
		gd.records_total = DmNumRecords(CurrentDB);

		// create/open StatusDB
		if ((err = DmCreateDatabase(0,"Thai Dict-Stat",DictStatCreator,DictType,false))) {
			FrmCustomAlert(AlertForm,"Cannot write to memory.",NULL,NULL);
			quit = 1;
			return 1;
		}
		StatusDB = DmOpenDatabaseByTypeCreator(DictType,DictStatCreator,dmModeReadWrite);
		// create status
		rechandle = DmNewRecord(StatusDB, 0, sizeof(GlobalData));
		if (rechandle==0) {
			FrmCustomAlert(AlertForm,"Not enough memory.",NULL,NULL);
			quit = 1;
			return 1;
		}
		recptr = MemHandleLock(rechandle);
		DmWrite(recptr, 0, &gd, sizeof(GlobalData));
		MemHandleUnlock(rechandle);
		DmReleaseRecord(StatusDB, 0, false);
		ReadDatabaseInfo();
		err = LoadWordRecord(gd.current_word_record, false);
		if (err) return 1;
		gd.word_ptr = gd.word_buf;
		FrmShowObject(frm, butClose);
		FrmSetTitle(frm, oldtitle);
		FrmSetActiveForm(frm);
		FrmDoDialog(frm);
		FrmDeleteForm(frm);
		LoadData2Array();
	}

	FrmGotoForm(DictMain);

	return 0;
}

void EventLoop(void)
{
	Err err;
	int formID;
	FormPtr form;
	EventType event;

	do
	{

		EvtGetEvent(&event, -1);  // Wait forever for an event,

		if (SysHandleEvent(&event)) continue;
		if (MenuHandleEvent((void *)0, &event, &err)) continue;

		if (event.eType == frmLoadEvent)  // If the event was that the form was just loaded,
		{
			formID = event.data.frmLoad.formID;	// Get the Id of the form loaded,
			form = FrmInitForm(formID);			// initialise the form
			FrmSetActiveForm(form);				// make the form active.
			switch (formID)
			{
				case DictMain:
				 mainform = form;
				 FrmSetEventHandler(form, (FormEventHandlerPtr) FormDictMainHandler);
				 // Set further events to be handled by our own function, FormHandler
				break;

				case MeanView:
				 FrmSetEventHandler(form, (FormEventHandlerPtr) FormMeanViewHandler);
				 // Set further events to be handled by our own function, FormHandler
				break;
			}
		}
		FrmDispatchEvent(&event); // Deal with the event, whatever it was.
		if (quit) break;

	} while(event.eType != appStopEvent);
	// If someone presses a hardware button, that will run another app, but will
	// signal this app (dict) to Stop ( with AppStopEvent ).

}

void StopApplication(void)
{
	/* ListPreferences listPreferences; */

	/* Remember the state of the application for the next startup */
	/* PrefGetAppPreferencesV10( ListType, ListVersion, &listPreferences, sizeof(ListPreferences));
	StrCopy( listPreferences.lastDBName, ListDBName );
	PrefSetAppPreferencesV10( ListType, ListVersion, &listPreferences, sizeof( ListPreferences )); */

	// save status
	MemPtr recptr;

	if (quit!=1) {
		recptr = MemHandleLock(DmGetRecord(StatusDB, 0));
		DmWrite(recptr, 0, &gd, sizeof(GlobalData));
		MemPtrUnlock(recptr);
		DmReleaseRecord(StatusDB, 0, false);
	}

	/* Close all the open forms */
	FrmCloseAllForms();

	/* Close the database */
	CloseDatabase();
}

Boolean FormDictMainHandler(EventPtr event)
{
	FormPtr form = FrmGetActiveForm();
	Err err;
	Boolean handled = false;
	TablePtr  table;
	FieldPtr  field;	
	UInt16 i, index;
	ControlPtr ctl=NULL;
	RectangleType bounds;

	switch (event->eType)   // Find out what sort of event it was.
	{
	case frmOpenEvent:

		/* Initialize the word list table */
		table = FrmGetObjectPtr( form, FrmGetObjectIndex( form, WordTable ));

		for( i = 0; i < wordTableRows; i++ ) {
			TblSetItemStyle( table, i, 0, customTableItem );
			TblSetRowUsable( table, i, false );
		}

		TblSetColumnUsable( table, 0, true );

		/* Set the callback routine that will draw the records */
		TblSetCustomDrawProcedure(table, 0, (TableDrawItemFuncPtr) &WordTableDrawRecord );

		LoadWordTable();

		FrmDrawForm( form );

		/* Give the quick lookup text the focus */
		FrmSetFocus( form, FrmGetObjectIndex( form, InputWord ));
		field = FrmGetObjectPtr( form, FrmGetObjectIndex( form, InputWord ));
		FldGrabFocus( field );

		handled = true;
		break;

	case penDownEvent:  // A control button was pressed and released.
		index = FrmGetObjectIndex(form, ClearGadget);
		FrmGetObjectBounds(form, index, &bounds);
		if (RctPtInRectangle(event->screenX, event->screenY, &bounds)) {
			field = FrmGetObjectPtr( form, FrmGetObjectIndex( form, InputWord ) );
			i = FldGetTextLength( field );
			FldDelete(field, 0, i);
			handled = true;
		}
		else {
			index = FrmGetObjectIndex(form, FindGadget);
			FrmGetObjectBounds(form, index, &bounds);
			if (RctPtInRectangle(event->screenX, event->screenY, &bounds))
				handled = QuickScroll( event );
		}
		break;

	case menuEvent:  // If it was a menu event....
		switch (event->data.menu.itemID)	// Find which menu item it was, and display the right form.
	    {
		case OptionsAbout:
		 	ShowMessageDialog(AboutForm);
			break;
		default:
			FrmCustomAlert(AlertForm,"No such options.",NULL,NULL);
			break;
	    }
		 handled = true;
	    break;

	case ctlRepeatEvent:
		/* Do not set handle = true for repeat events so they will repeat automatically */ 
		table = FrmGetObjectPtr( form, FrmGetObjectIndex( form, WordTable ));
		switch( event->data.ctlRepeat.controlID ) {
			case UpButton:
				/* Scroll Up: Backup either mainTableRows - 1 elements, or as far as we can */
				ctl = FrmGetObjectPtr( form, FrmGetObjectIndex( form, UpButton ));
				if( ctl->attr.visible && CtlEnabled( ctl ) ) {
					gd.current_word_record = gd.word_table[0].word_record_number;
					err = LoadWordRecord(gd.current_word_record, true);
					if (err) return true;
					GoBack( wordTableRows, gd.word_table[0].ptr_in_record);
				}
				break;
							
			case DownButton:
				/* Scroll Down: make the top visible record the bottom record currently displayed */
				ctl = FrmGetObjectPtr( form, FrmGetObjectIndex( form, DownButton ));
				if( ctl->attr.visible && CtlEnabled( ctl ) ) {
					gd.current_word_record = gd.word_table[gd.last_idx].word_record_number;
					gd.word_ptr = gd.word_table[gd.last_idx].ptr_in_record;
				}
				break;
		}
		if( ctl->attr.visible && CtlEnabled( ctl ) ) { 
			LoadData2Array();
			LoadWordTable();
			TblEraseTable( table );
			TblDrawTable( table );
		}
		break;

	case keyDownEvent: 
		table = FrmGetObjectPtr( form, FrmGetObjectIndex( form, WordTable ));
		ctl = NULL;

		switch( event->data.keyDown.chr ) {
			case pageUpChr:
				/* Scroll Up: Backup either mainTableRows - 1 elements, or as far as we can */
				ctl = FrmGetObjectPtr( form, FrmGetObjectIndex( form, UpButton ));
				if( ctl->attr.visible && CtlEnabled( ctl ) ) {
					gd.current_word_record = gd.word_table[0].word_record_number;
					err = LoadWordRecord(gd.current_word_record, true);
					if (err) return true;
					GoBack( wordTableRows, gd.word_table[0].ptr_in_record);
				}
				break;

			case pageDownChr:
				/* Scroll Down: make the top visible record the bottom record currently displayed */
				ctl = FrmGetObjectPtr( form, FrmGetObjectIndex( form, DownButton ));
				if( ctl->attr.visible && CtlEnabled( ctl ) ) {
					gd.current_word_record = gd.word_table[gd.last_idx].word_record_number;
					gd.word_ptr = gd.word_table[gd.last_idx].ptr_in_record;
				}
				break;

			case chrLineFeed:
				handled = QuickScroll( event );
				break;
		}

		if( ctl && ctl->attr.visible && CtlEnabled( ctl ) ) { 
			LoadData2Array();
			LoadWordTable();
			TblEraseTable( table );
			TblDrawTable( table );
		}

		break;

	case tblSelectEvent:
		gd.select_idx = TblGetRowID( event->data.tblSelect.pTable, event->data.tblSelect.row );
		gd.select_word_record = gd.word_table[gd.select_idx].word_record_number;
		gd.select_word_ptr = gd.word_table[gd.select_idx].ptr_in_record;
		FrmPopupForm(MeanView);
		handled = true;
		break;

	default:
		break;
  }
  return handled;
}


// Show a message, if necessary.
void ShowMessageDialog(UInt16 fid)
{
	FormPtr form;
	form = FrmInitForm(fid);
	if (form)
	{
		FrmDoDialog(form);
		FrmDeleteForm(form);
	}
}

void ReadDatabaseInfo()
{
	LocalID dbID;
	UInt16 cardNo; 
	LocalID appInfoID;
	DictAppInfoType *appInfo;

	/* Read the name from the current database */
	DmOpenDatabaseInfo( CurrentDB, &dbID, NULL, NULL, &cardNo, NULL );
	DmDatabaseInfo( cardNo, dbID, DictDBName, NULL, NULL, NULL, NULL, NULL,
					NULL, &appInfoID, NULL, NULL, NULL );
	appInfo = MemLocalIDToLockedPtr( appInfoID, cardNo );

	gd.start_word_record = 0;
	gd.end_word_record = appInfo->WordRecTotal - 1;
	gd.start_mean_record = appInfo->WordRecTotal;
	gd.end_mean_record = appInfo->WordRecTotal + appInfo->MeanRecTotal - 1;

	gd.current_word_record = gd.start_word_record;
	gd.current_mean_record = gd.start_mean_record;

	MemPtrUnlock( appInfo );
}

/* Close the CurrrentDB and update database specific info (current category) */
void CloseDatabase()
{
	LocalID dbID;
	UInt16 cardNo;
	LocalID appInfoID;
	DictAppInfoType *appInfo;

	DmOpenDatabaseInfo( CurrentDB, &dbID, NULL, NULL, &cardNo, NULL );
	DmDatabaseInfo( cardNo, dbID, NULL, NULL, NULL, NULL, NULL, NULL, 
					NULL, &appInfoID, NULL, NULL, NULL );
	appInfo = MemLocalIDToLockedPtr( appInfoID, cardNo );

	MemPtrUnlock( appInfo );

	if (quit!=1) DmCloseDatabase(StatusDB);

	DmCloseDatabase( CurrentDB );
}

/* This draws the list view of the database entries.  
In the future we will have three(?) different draw policies based
on the database info.  For now we will always use a single draw
policy */
void WordTableDrawRecord(MemPtr table, UInt16 row, UInt16 column, RectanglePtr bounds )
{
	//MemHandle recordHandle;
	//DictDBHeader *record;
	//UInt16 recordNumber;

	//recordNumber = TblGetRowID( table, row );  /* the row ID contains the record # */
	// recordHandle = DmQueryRecord( CurrentDB, recordNumber );

	/* If the DmQueryRecord fails, don't use the pointer */
	//if( recordHandle ) {
		//record = MemHandleLock( recordHandle );
		DictDrawEntry( row, bounds );
		//MemHandleUnlock( recordHandle  ); /*  remember to unlock what we lock down */
	//}
}


/*****************************
 *	Load word data into array *
 *****************************/
UInt16 LoadData2Array()
{
	Err err;

	char *data;
	int tabIndex, sz=0;
	Boolean endloop, wordonly;
	UInt16  first_word_record;

	gd.lastpage = false;

	data = gd.word_ptr;
	first_word_record = gd.current_word_record;
	
	tabIndex = 0;

	endloop = true;
	wordonly = true;

	MemSet(gd.word_table, wordTableRows*sizeof(DictTableInfo), 0);

	while( endloop ) {
		switch( *data ) {
		case DELIMITER: /* if match '#' */
			gd.word_table[tabIndex].word_record_number = gd.current_word_record;
			gd.word_table[tabIndex].ptr_in_record = data;
			data++;
			sz = 0;
			break;
		case TYPE_DELIMITER: /* if match ':' */
			MemMove(gd.word_table[tabIndex].word,(data-sz),sz);
			wordonly = false;
			data++;
			sz = 0;
			break;
		case ADDR_DELIMITER: /* if match '@' */
			if (wordonly) {
				MemMove(gd.word_table[tabIndex].word,(data-sz),sz);
			} else {
				MemMove(gd.word_table[tabIndex].word_type,(data-sz),sz);
			}
			MemMove(&gd.word_table[tabIndex].mean_record_number,(data+1),1);
			gd.word_table[tabIndex].mean_record_number += gd.start_mean_record;
			MemMove(&gd.word_table[tabIndex].mean_offset,(data+2),2);

			wordonly = true;
			data+=4;
			sz = 0;
			tabIndex++;
			if (tabIndex >= wordTableRows) {
				endloop = false;
			}
			break;
		case '\0':
			if (gd.current_word_record == gd.end_word_record) {
				endloop = false;
				gd.lastpage = true;
			} else {
				gd.current_word_record++;
				err = LoadWordRecord(gd.current_word_record, true);
				if (err) return -1;
				data = gd.word_buf;
				wordonly = true;
			}
			break;
		default:
			data++;
			sz++;
		}
	}
	gd.last_idx = tabIndex-1;

	return 0;
}

/* This routine is used to update the main form table.	It is used when:
- intializing the table
- scrolling

Answers the row to highlight if the current record is shown, if the current
row is not shown it answers -1.
*/
UInt16 LoadWordTable()
{
	FormPtr form;
	TablePtr table;
	UInt16 i;
	UInt16 select;
	UInt16 upIndex, downIndex;
	Boolean scrollUp, scrollDown;

	form = FrmGetActiveForm();
	table = FrmGetObjectPtr( form, FrmGetObjectIndex( form, WordTable ));

	scrollUp = true;
	if ((gd.word_ptr == gd.word_buf) && (gd.current_word_record == gd.start_word_record)) scrollUp = false;

	scrollDown = true;
	if (gd.lastpage) scrollDown = false;

	select = -1;
	/* For every table row, fetch the associated record */
	for( i = 0; i < wordTableRows; i++ ) {
		if (gd.word_table[i].word[0] == 0) {
			/* Make the row non-usable */
			TblSetRowUsable( table, i, false );
		} else {
			/* Make the row usable */
			TblSetRowUsable( table, i, true );
			/* Mark it invalid so it will be redrawn */
			TblMarkRowInvalid( table, i );
			/*	Set the row id to be the record number */
			TblSetRowID( table, i, i );
		}
	}

	TblUnhighlightSelection( table );
	
	upIndex = FrmGetObjectIndex( form, UpButton );
	downIndex = FrmGetObjectIndex( form, DownButton );

	FrmUpdateScrollers( form, upIndex, downIndex, scrollUp, scrollDown ); 

	return( select );
}


void DictDrawEntry( UInt16 row, RectanglePtr bounds)
{
	char *string;
	Int16 x;
	UInt32 length;
	Boolean needDots = false;
	Boolean done = false;
	Int16 remaining, newLength;
	UInt16 columnWidth;

	FntSetFont( stdFont );
  
	columnWidth = bounds->extent.x;

	if( gd.word_table[row].word != 0 ) {
		string = gd.word_table[row].word;
	} else {
		string = emptyRecordLabel;
	}

	x = bounds->topLeft.x;
	length = StrLen( string );

	/* Restrict the first field to a maximum of 110 pixels */
	if( FntCharsWidth( string, length ) > 110 ) {
		remaining = 110;
		newLength = length;
		FntCharsInWidth( string, &remaining, &newLength, &needDots );
		length = newLength;
		needDots = true;
	}
	WinDrawChars( string, length, x, bounds->topLeft.y );
  
	/* If We truncated the string at all, put some dots */
	x += FntCharsWidth( string, length );
	if( needDots ) {
		WinDrawChars( "..", 2, x, bounds->topLeft.y );
		x += FntCharsWidth( "..", 2 );
		needDots = false;
	}
  
	done = gd.word_table[row].word_type == 0;
	if( !done ) {    
		/* Calculate the amount of room remaining for the second field.
		This is the minimum X value we can use for the next field */
		remaining = (columnWidth - 4) - x; /* note: fudge factor of 4 pixels */
    
		string = gd.word_table[row].word_type;

		length = StrLen( string );
    
		/* Restrict the second field to the remaining pixel size */
		if((x = FntCharsWidth( string, length )) > remaining ) {
			remaining -= FntCharsWidth( "..", 2 );

			newLength = length;
			FntCharsInWidth( string, &remaining, &newLength, &needDots );
			length = newLength;
			needDots = true;
			x = FntCharsWidth( string, length );
		}
		x = columnWidth - x; 
    
		if( needDots ) {
			x -= FntCharsWidth( "..", 2 );
		}
    
		WinDrawChars( string, length, x, bounds->topLeft.y );
		if( needDots ) {
			x += FntCharsWidth( string, length );
			WinDrawChars( "..", 2, x, bounds->topLeft.y );
		}
	}
}

Err LoadWordRecord(int recordNo, Boolean showstat)
{
	MemHandle rec_handle;

	if ((recordNo >= gd.start_word_record) && (recordNo <= gd.end_word_record))
	{
		decompress_record(recordNo, showstat);
		if (quit) return 1;
		rec_handle = DmQueryRecord(StatusDB, 1);

		gd.word_buf = MemHandleLock(rec_handle);
		gd.word_buf_size = MemHandleSize(rec_handle);

		MemHandleUnlock(rec_handle);
	}
	return 0;
}

Err LoadMeanRecord(int recordNo, Boolean showstat)
{
	MemHandle rec_handle;

	if ((recordNo >= gd.start_mean_record) && (recordNo <= gd.end_mean_record))
	{
		decompress_record(recordNo, showstat);
		if (quit) return 1;
		rec_handle = DmQueryRecord(StatusDB, 2);

		gd.mean_buf = MemHandleLock(rec_handle);
		gd.mean_buf_size = MemHandleSize(rec_handle);

		MemHandleUnlock(rec_handle);
	}
	return 0;
}

Boolean QuickScroll( EventPtr event ) {
	TablePtr table;
	FormPtr form;
	FieldPtr field;
	UInt16 fieldIndex;
	char *alphaSrc='\0';
	char alphaDst[64];
	UInt16 quickTextLength;
	/* Some random key has been input -- if there exists a record
	in the current category whose first letter in the display
	style (ie: sort order) matches, then scroll such that this
	record is visible (top of the screen?) */
	char *data, *end_data;
	Int16 sz=0, i, comp_res, record_no=-1;
	Boolean endloop, wordonly;
	RectangleType rp;
	Err err;

	/* Update the quick scroll text line */
	form = FrmGetActiveForm();
	table = FrmGetObjectPtr( form, FrmGetObjectIndex( form, WordTable ));
	fieldIndex = FrmGetObjectIndex( form, InputWord );
	FrmSetFocus( form, fieldIndex );
	field = FrmGetObjectPtr( form, fieldIndex );

	quickTextLength = FldGetTextLength( field );

	/* if the text is of zero length, unhighlight any entry in the table */
	if( quickTextLength == 0 ) {
		TblUnhighlightSelection( table );
		return true;
	}

	if( FldHandleEvent( field, event ) || (event->eType == ctlSelectEvent) ) {
		StrToLower( alphaSrc, FldGetTextPtr( field ) );

		/* compare first word in record */
		comp_res = -1;
		// optimizing the search
		if (alphaSrc[0]-'a'<0) {
			FldDelete(field, 0, quickTextLength);
			return true;
		}
		else {
			// hardwired word lookup
			if ((comp_res=StrCompare(alphaSrc, "zygote"))>0) record_no=-1;
			else if ((comp_res=StrCompare(alphaSrc, "unaided"))>=0) record_no=7;
			else if ((comp_res=StrCompare(alphaSrc, "sherry"))>=0) record_no=6;
			else if ((comp_res=StrCompare(alphaSrc, "pluperfect"))>=0) record_no=5;
			else if ((comp_res=StrCompare(alphaSrc, "materially"))>=0) record_no=4;
			else if ((comp_res=StrCompare(alphaSrc, "hippopotamus"))>=0) record_no=3;
			else if ((comp_res=StrCompare(alphaSrc, "dunmow flitch"))>=0) record_no=2;
			else if ((comp_res=StrCompare(alphaSrc, "cassock"))>=0) record_no=1;
			else if ((comp_res=StrCompare(alphaSrc, "a"))>=0) record_no=0;

			if (record_no==-1) {
				FldDelete(field, 0, quickTextLength);
				return true;
			}
		}

		RctSetRectangle(&rp, 40, 70, 80, 20);
		WinEraseRectangle(&rp, 0);
		WinDrawRectangleFrame(dialogFrame, &rp);
		FntSetFont( boldFont );
		WinDrawChars(" Searching... ",14,50,75);
		FntSetFont( stdFont );

		err = LoadWordRecord(record_no, false);
		if (err) return false;
		gd.current_word_record = record_no;

		if (comp_res==0) { /* match first word in record */

			if (record_no==0) {
				i = 1;
			} else {
				i = (wordTableRows+1)/2; /* set found word at center screen */
			}
			gd.word_ptr = gd.word_buf;
			GoBack( i, gd.word_ptr);

		} else {

			/* find word in current record */
			data = gd.word_buf;
			end_data = gd.word_buf+gd.word_buf_size;
			comp_res = 1;
			gd.word_ptr = data;
			while ((*data == DELIMITER) && (comp_res>0)) {
				endloop = true;
				wordonly = true;
				MemSet(alphaDst, 64, 0);
				while( endloop ) {
					switch( *data ) {
					case DELIMITER: /* if match '#' */
						gd.word_ptr = data;
						data++;
						if (*data == '-' || *data == '\'') { data++; }
						sz = 0;
					break;
					case TYPE_DELIMITER: /* if match ':' */
						MemMove(alphaDst,(data-sz),sz);
						StrToLower( alphaDst, alphaDst );
						wordonly = false;
						data++;
						sz = 0;
						break;
					case ADDR_DELIMITER: /* if match '@' */
						if ( wordonly ) {
							MemMove(alphaDst,(data-sz),sz);
							StrToLower( alphaDst, alphaDst );
						}
						endloop = false;
						data += 4;
						sz = 0;
						break;
					default:
						data++;
						sz++;
					}
				}
				comp_res = StrCompare(alphaSrc, alphaDst);
			}

			if ( (record_no==7) && (*data != DELIMITER) ){
				i = wordTableRows-1;
			} else {
				i = (wordTableRows+1)/2; /* set found word at center screen */
			}
			GoBack( i, gd.word_ptr);

		}

		LoadData2Array();
		LoadWordTable();
		TblEraseTable( table );
		TblDrawTable( table );
		if (i > 0)
		{
			TblSelectItem( table, i-1, 0 );	
			gd.select_idx = i-1;
		}

		return true;
	}
	return false;
}

Err GoBack(int item, char *start) {
	Err err;
	Boolean endloop;
	int linecount;
	char *data;

	endloop = true;
	linecount = item;
	data = start;
	while( endloop ) {
		if (data == gd.word_buf) {
			if (gd.current_word_record == gd.start_word_record) {
				data = gd.word_buf;
				endloop = false;
			} else {
				linecount--;
				gd.current_word_record--;
				err = LoadWordRecord(gd.current_word_record, true);
				if (err) return 1;
				data = gd.word_buf+gd.word_buf_size - 4;
			}
		} else {
			switch( *data ) {
				case DELIMITER:
					linecount--;
					if (linecount <= 0) {
						endloop = false;
					} else {
						data -= 4;
					}
				break;
				default:
					data--;
			}
		}
	}
	gd.word_ptr = data;
	return 0;
}

/******************************************
 *	Decompress record #recno					*
 *	Input:	record number (in CurrentDB)	*
 *          show status                   *
 *	Output:	Status DB at						*
 *				= record #1 (for word)			*
 *			or	= record #2 (for mean)			*
 ******************************************/
Err decompress_record(UInt16 recno, Boolean showstat) {
	Err err = 0;
	UInt16 index = dmMaxRecordIndex, status_recno;
	MemHandle rec, drec;
	MemPtr prec, pdrec;
	z_stream ds;
	UInt32 len, count=0, dif=0;
	RectangleType rp;

	// create/lock a decompress record
	if ((recno >= gd.start_word_record) && (recno <= gd.end_word_record)) {
		status_recno = 1;
		// if already loaded, just quit
		if (recno == gd.de_word_record) return 0;
		if (gd.de_word_record!=255)
			DmRemoveRecord(StatusDB, status_recno);
	}
	else {
		status_recno = 2;
		// if already loaded, just quit
		if (recno == gd.de_mean_record) return 0;
		if (gd.de_mean_record!=255)
			DmRemoveRecord(StatusDB, status_recno);
	}

	// reserve mem for output
	drec = DmNewRecord(StatusDB, &index, MAX_RECORD_SIZE);
	if (drec==0) {
		FrmCustomAlert(AlertForm,"Not enough memory.",NULL,NULL);
		if ((recno >= gd.start_word_record) && (recno <= gd.end_word_record))
			quit = 1;
		else
			quit = 2;
		return err;
	}
	pdrec = MemHandleLock(drec);	

	if ((recno >= gd.start_word_record) && (recno <= gd.end_word_record))
		gd.de_word_record = recno;
	else gd.de_mean_record = recno;

	// lock/find len of the record
	rec = DmGetRecord(CurrentDB, recno);
	prec = MemHandleLock(rec);
	len = MemHandleSize(rec);

	if (showstat) {
		RctSetRectangle(&rp, 40, 70, 80, 20);
		WinEraseRectangle(&rp, 0);
		WinDrawRectangleFrame(dialogFrame, &rp);
		FntSetFont( boldFont );
		WinDrawChars("Decompressing",13,43,75);
		FntSetFont( stdFont );
	}

	// initial inflate
	ds.zalloc = (alloc_func)0;
	ds.zfree = (free_func)0;
	ds.opaque = (voidpf)0;
	ds.next_in = prec;
	ds.next_out = buffer;
	ds.avail_in = len;
	ds.avail_out = DE_BUFFER_SIZE;
	err = inflateInit(&ds);
	buffer[0]='\0';

	// limit at MAX_SIZE
	while (ds.total_in < len && ds.total_out < MAX_RECORD_SIZE && err != Z_STREAM_END) {
		err = inflate(&ds, Z_NO_FLUSH);
		dif = DE_BUFFER_SIZE - ds.avail_out;
		DmWrite(pdrec, count, buffer, dif);
		count += dif;
		ds.next_out = buffer;
		ds.avail_out = DE_BUFFER_SIZE;
	}

	err = inflateEnd(&ds);

	// Unlock/resize decompress (output) record
	MemHandleUnlock(drec);
	DmReleaseRecord(StatusDB, index, false);
	DmResizeRecord(StatusDB, index, count);
	if (index!=status_recno) DmMoveRecord(StatusDB, index, status_recno);

	// Unlock data record
	MemHandleUnlock(rec);
	DmReleaseRecord(CurrentDB, recno, false);

	if (showstat) {
		WinEraseRectangleFrame(dialogFrame, &rp);
		WinEraseRectangle(&rp, 0);
	}

	return 0;
}
