/* meanview.c
 */

/* Includes required */
#include <PalmOS.h>

#include "dictrc.h"
#include "dict.h"

static void UpdateScrollBar ()
{
	Int16 scrollPos;
	Int16 textHeight;
	Int16 fieldHeight;
	Int16 maxValue;
	FormPtr form = FrmGetActiveForm();
	FieldPtr fld;
	ScrollBarPtr bar;

	fld = FrmGetObjectPtr( form, FrmGetObjectIndex( form, MeanField ));
	bar = FrmGetObjectPtr( form, FrmGetObjectIndex( form, MeanScrollBar ));
	
	// get the values necessary to update the scroll bar.
	FldGetScrollValues (fld, &scrollPos, &textHeight,  &fieldHeight);

	if (textHeight > fieldHeight)
		maxValue = textHeight - fieldHeight;
	else if (scrollPos)
		maxValue = scrollPos;
	else
		maxValue = 0;

	SclSetScrollBar (bar, scrollPos, 0, maxValue, fieldHeight-1);
}

static void PageScroll (WinDirectionType direction)
{
	Int16 value;
	Int16 min;
	Int16 max;
	Int16 pageSize;
	Int16 linesToScroll;
	FormPtr form = FrmGetActiveForm();
	FieldPtr fld;
	ScrollBarPtr bar;

	fld = FrmGetObjectPtr( form, FrmGetObjectIndex( form, MeanField ));
	
	if (FldScrollable (fld, direction)) {
		linesToScroll = FldGetVisibleLines (fld) - 1;
		FldScrollField (fld, linesToScroll, direction);

		// Update the scroll bar.
		bar = FrmGetObjectPtr( form, FrmGetObjectIndex( form, MeanScrollBar ));
		SclGetScrollBar (bar, &value, &min, &max, &pageSize);

		if (direction == winUp)
			value -= linesToScroll;
		else
			value += linesToScroll;
		
		SclSetScrollBar (bar, value, min, max, pageSize);
	}
}

static void Scroll (Int16 linesToScroll)
{
	Int16			blankLines;
	Int16			min;
	Int16			max;
	Int16			value;
	Int16			pageSize;
	FormPtr form = FrmGetActiveForm();
	FieldPtr			fld;
	ScrollBarPtr	bar;
	
	fld = FrmGetObjectPtr( form, FrmGetObjectIndex( form, MeanField ));

	if (linesToScroll < 0) {
		blankLines = FldGetNumberOfBlankLines (fld);
		FldScrollField (fld, -linesToScroll, winUp);
		
		// If there were blank lines visible at the end of the field
		// then we need to update the scroll bar.
		if (blankLines) {
			// Update the scroll bar.
			bar = FrmGetObjectPtr( form, FrmGetObjectIndex( form, MeanScrollBar ));
			SclGetScrollBar (bar, &value, &min, &max, &pageSize);
			if (blankLines > -linesToScroll)
				max += linesToScroll;
			else
				max -= blankLines;
			SclSetScrollBar (bar, value, min, max, pageSize);
		}
	} else if (linesToScroll > 0)
		FldScrollField (fld, linesToScroll, winDown);
}

Boolean FormMeanViewHandler(EventPtr event)
{
	Boolean handled = false;
	FormPtr form = FrmGetActiveForm();
	FieldPtr fld;
	RectangleType rect;
	char *string;
	UInt16 stringLength;
	MemHandle textHandle;
	MemPtr textPtr;
	Err err;

	RctSetRectangle(&rect,0,16,160,128);
	switch( event->eType )
	{
	case keyDownEvent: 
		if (event->data.keyDown.chr == chrPageUp)	 {
			PageScroll(winUp);
			handled = true;
		} else if (event->data.keyDown.chr == chrPageDown) {
			PageScroll(winDown);
			handled = true;
		}
		break;

	case sclRepeatEvent:
		Scroll (event->data.sclRepeat.newValue - event->data.sclRepeat.value);
		break;

	case ctlSelectEvent:
		if( event->data.ctlSelect.controlID == MeanCloseButton ) {
			TablePtr  table;
			FieldPtr  field;
			FrmReturnToForm(DictMain);
			// focus the field
			form = FrmGetActiveForm();
			field = FrmGetObjectPtr( form, FrmGetObjectIndex( form, InputWord ));
			FldGrabFocus( field );
			FrmSetFocus(form, FrmGetObjectIndex( form, InputWord ));
			// redraw table
			table = FrmGetObjectPtr( form, FrmGetObjectIndex( form, WordTable ));
			TblEraseTable( table );
			TblDrawTable( table );
			TblSelectItem( table, gd.select_idx, 0 );
			handled = true;
		}
		break;

	case ctlRepeatEvent:

	case frmOpenEvent:

		fld = FrmGetObjectPtr( form, FrmGetObjectIndex( form, MeanField ));

		err = LoadMeanRecord(gd.word_table[gd.select_idx].mean_record_number, true);
		if (err) return true;

		if ( gd.word_table[gd.select_idx].mean_offset == 0xffff ) {
			stringLength = StrLen( gd.word_table[gd.select_idx].word_type ) + 1;
			textHandle = MemHandleNew( stringLength );
			if( textHandle == NULL ) {
				FrmCustomAlert(AlertForm,"Not enough memory.",NULL,NULL);
				return true;
			}
			StrCopy( MemHandleLock( textHandle ), gd.word_table[gd.select_idx].word_type );
			MemHandleUnlock( textHandle );
		} else {
			string = gd.mean_buf + gd.word_table[gd.select_idx].mean_offset;
			stringLength = StrLen( gd.word_table[gd.select_idx].word_type ) + StrLen( string ) + 2;
			textHandle = MemHandleNew( stringLength );
			if( textHandle == NULL ) {
				FrmCustomAlert(AlertForm,"Not enough memory.",NULL,NULL);
				return true;
			}
			textPtr = MemHandleLock( textHandle );
			if ( StrLen( gd.word_table[gd.select_idx].word_type ) > 0 ) {
				StrCopy( textPtr, gd.word_table[gd.select_idx].word_type );
				StrCat( textPtr, "\n" );
				StrCat( textPtr, string );
			} else {
				StrCopy( textPtr, string );
			}
			MemHandleUnlock( textHandle );
		}
		FldSetText( fld, textHandle, 0, stringLength );

		/* Update scrollers */
		UpdateScrollBar();

		FrmDrawForm( form );

		if ( StrLen(gd.word_table[gd.select_idx].word) <= 20 )
			StrCopy( ViewTitle, gd.word_table[gd.select_idx].word );
		else
			StrNCopy( ViewTitle, gd.word_table[gd.select_idx].word, 20 );

		FntSetFont( largeBoldFont );
		WinDrawChars( ViewTitle, StrLen(ViewTitle), 1, 0 );
		WinDrawLine(0,16,160,16);
		FntSetFont( stdFont );

		handled = true;
		break;

	default:
		break;
	}

	return( handled );
}
