// vi: set ts=3
#include <PalmOS.h>
#include "zlib.h"

#define	BUFFER_SIZE		1024		// 1 k block
#define	MAX_SIZE			65505		// about 64k limit

/*
 *   Compress Record #recno
 *   Input:		database ref, record number
 *   Output:	compressed record on database ref @ record number
 */
UInt32 compress_record(DmOpenRef db, UInt16 recno) {
	Err err;
	UInt16 index = dmMaxRecordIndex;
	MemHandle rec, crec;
	MemPtr prec, pcrec;
	z_stream cs;
	UInt32 len, count=0, dif=0;
	char buffer[BUFFER_SIZE];

	// read the original data from record number #recno
	rec = DmGetRecord(db, recno);
	prec = MemHandleLock(rec);
	len = MemHandleSize(rec);

	// create a new record for compressed data
	crec = DmNewRecord(db, &index, len);
	pcrec = MemHandleLock(crec);

	// initial deflate
	cs.zalloc = (alloc_func)0;
	cs.zfree = (free_func)0;
	cs.opaque = (voidpf)0;
	cs.next_in = prec;
	cs.next_out = buffer;
	cs.avail_in = len;
	cs.avail_out = BUFFER_SIZE;
	err = deflateInit(&cs, Z_DEFAULT_COMPRESSION);

	// wow.. is there any case that compressed data larger than uncompressed ?
	// guess no.. so set limit to uncompressed size
	while (cs.avail_in != 0 && cs.total_out < len) {
		deflate(&cs, Z_NO_FLUSH);
		dif = BUFFER_SIZE - cs.avail_out;
		DmWrite(pcrec, count, buffer, dif);
		count += dif;
		cs.next_out = buffer;
		cs.avail_out = BUFFER_SIZE;
	}

	while (err!=Z_STREAM_END && cs.total_out < len) {
		cs.avail_in = 0;
		cs.avail_out = BUFFER_SIZE;
		cs.next_out = buffer;
		err = deflate(&cs, Z_FINISH);
		dif = BUFFER_SIZE - cs.avail_out;
		DmWrite(pcrec, count, buffer, dif);
		count += dif;
		cs.next_out = buffer;
	}

	deflateEnd(&cs);
	MemHandleUnlock(crec);
	DmReleaseRecord(db, index, true);
	DmResizeRecord(db, index, count);

	MemHandleUnlock(rec);
	DmReleaseRecord(db, recno, true);

	// swap the record
	DmRemoveRecord(db, recno);
	if (recno!=index-1) DmMoveRecord(db, index-1, recno);

	return count;
}

UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
	if (cmd==sysAppLaunchCmdNormalLaunch) {
		EventType e;
		UInt32 size=0;
		UInt16 i;
		char str[40], val[10];
		DmOpenRef db;
		LocalID dbid;
		UInt16 cardno = 0;
		UInt32 mod = 0, now = 0;

		// open database
		db = DmOpenDatabaseByTypeCreator('DATA', 'TPdb', dmModeReadWrite);

		StrCopy(str, "Start compression.");
		WinDrawChars(str, StrLen(str), 0, 10);

		// Compress record # [0-43] of DB , now is 44 records
		for (i=0;i<44;i++) {
			StrCopy(str, "Count: [");
			StrIToA(val, i);
			StrCat(str, val);
			StrCat(str, "]");
			WinDrawChars(str, StrLen(str), 0, 20);
			size += compress_record(db, i);
			StrCopy(str, "Total compressed size: ");
			StrIToA(val, size);
			StrCat(str, val);
			WinDrawChars(str, StrLen(str), 0, 30);
		}
		StrCopy(str, "Done. You may export the pdb now.");
		WinDrawChars(str, StrLen(str), 0, 40);

		now = TimGetSeconds();
		DmOpenDatabaseInfo(db,&dbid,NULL,NULL,&cardno,NULL);
		DmSetDatabaseInfo(cardno,dbid,NULL,NULL,NULL,&now,&now,&now,&mod,NULL,NULL,NULL,NULL);

		// close the database
		DmCloseDatabase(db);

		do {
			EvtGetEvent(&e, -1);
			if (SysHandleEvent(&e)) continue;
		} while (e.eType != appStopEvent);

		FrmCloseAllForms();
	}
	return 0;
}
